module AccessibleAutomata where

import Automata
import RandomState
import TabakovVardi
import Data.List
import Data.Ratio
import qualified Control.Monad.State as MS

--default (Int)

nufix :: Eq a => ([a] -> [a]) -> [a] -> [a]
nufix f s0 = let new = nub (f s0) in 
             if new \\ s0 == [] then new else nufix f new

data AccessType = Access | Coaccess | Biaccess deriving (Read,Show,Eq,Ord)

biAccessibleAutomaton :: AccessType -> Bool -> Int -> Int -> Rational -> Rational -> RandomState Automaton
--initTrans forces an outbound transition on each character from the start state.
--Takes sigma = number of characters, n = number of states, r = transition density, f = final state 
--density to generate a bi-accessible automaton.
biAccessibleAutomaton access initTrans sigma n r f = 
        do
          finStates <- getRandomSubset [0..n-1] (floor (f * (fromIntegral n)))
          accessibleTrans <- if access == Coaccess then return [] else accessibleEdges sigma n
          coaccessibleTrans <- if access == Access then return [] else coaccessibleEdges sigma n finStates accessibleTrans
          balancedTrans <- balance sigma n r (accessibleTrans ++ coaccessibleTrans)
          fullTrans <- remainingTrans sigma n r balancedTrans
          return (Automaton sigma [0..n-1] [0] fullTrans finStates)

accessibleEdges :: Int -> Int -> RandomState [(State,Int,State)]
--Generates the edges that would make an automaton accessible. sigma = number of chars, n = states
accessibleEdges sigma n = 
        edgeHelper sigma n [0] [1..n-1]
        where
          edgeHelper :: Int -> Int -> [State] -> [State] -> RandomState [(State,Int,State)]
          edgeHelper _ _ _ [] = return []
          edgeHelper sigma n vis uvis =
            do
              u <- getRandomElem vis
              v <- getRandomElem uvis
              c <- getRandomR (0,sigma-1)
              let newV = v : vis
              let newUV = delete v uvis
              restEdges <- edgeHelper sigma n newV newUV
              return $ (u,c,v) : restEdges

coaccessibleEdges :: Int -> Int -> [State] -> [(State,Int,State)] -> RandomState [(State,Int,State)]
coaccessibleEdges sigma n finStates accessTrans =
        edgeHelper sigma n (initVis) ([0..n-1] \\ initVis)
        where
          edgeHelper :: Int -> Int -> [State] -> [State] -> RandomState [(State,Int,State)]
          --identical to accessibleEdges version, but swaps u and v
          edgeHelper _ _ _ [] = return []
          edgeHelper sigma n vis uvis =
            do
              v <- getRandomElem vis
              u <- getRandomElem uvis
              c <- getRandomR (0,sigma-1)
              let newV = u : vis
              let newUV = delete u uvis
              restEdges <- edgeHelper sigma n newV newUV
              return $ (u,c,v) : restEdges

          initVis = nufix (inits accessTrans) finStates 
          --fixpoint operation to get connections more than 1 away from a final state
          inits :: [(State,Int,State)] -> [State] -> [State]
          inits trans vis = vis ++ [u | (u,a,v) <- trans, v `elem` vis]


balance :: Int -> Int -> Rational -> [(State,Int,State)]-> RandomState [(State,Int,State)]
--takes a list of transitions and makes sure there aren't too many of one kind of transition using n and k.
balance sigma n k es 
        | (numEdges*sigma) < length es = return es
        | otherwise =
          do
            let labeledCounts = map (\c -> (c, margin c)) [0..sigma-1]
                (unders, overs) = partition (\(c, mrg) -> mrg < 0) labeledCounts
            transToRemove <- (MS.liftM concat) $ sequence $ map (\(c, mrg) -> getRandomSubset (existTrans c) (-mrg)) unders
            let availChars = concat $ map (\(c, mrg) -> replicate mrg c) overs
            charsForChange <- getRandomMultisubset availChars (length transToRemove)
            let newTrans = zipWith (\(u,_,v) c -> (u,c,v)) transToRemove charsForChange
            return $ newTrans ++ (es \\ transToRemove)
        where
          margin c = numEdges - (length (existTrans c))
          numEdges = (floor ((fromIntegral n) * k))
          existTrans c = filter (\(_,char,_) -> char == c) es

remainingTrans :: Int -> Int -> Rational -> [(State,Int,State)]-> RandomState [(State, Int, State)] 
remainingTrans sigma n r trans = 
        do
           let graphOfChar c = genEdges n (count c) (existTrans c)
           trans <- sequence $ [MS.liftM (putChar c) (graphOfChar c) | c <- [0..sigma-1]]
           return $ concat trans 
        where
          count :: Int -> Int
          count c = numEdges - (length (existTrans c))
          existTrans c = remChar (filter (\(_,char,_) -> char == c) trans)
          numEdges = (floor ((fromIntegral n) * r))
          putChar c lst = [(u,c,v) | (u,v) <- lst]
          remChar lst = [(u,v) | (u,_,v) <- lst]
