{-# LANGUAGE ScopedTypeVariables #-}

module Main where

import Numeric
import System.Directory
import System.Environment
import System.IO
import System.Random
import Data.List
import Data.List.Split
import Data.Maybe
import Data.Char
import Data.Ratio
import Debug.Trace
import Data.Tuple.Curry
import Control.Monad.State as MS
--import System.Posix.Files
import System.Directory
import SmallWorldAutomata
import TabakovVardi
import ScaleFreeAutomata
import AccessibleAutomata
import Automata
import RandomState
import Kronecker
import VertexCopying
import Logit

import System.Console.GetOpt

--Stores every possible param that could be put in a config file; these should not all be filled out.
data AutomataParams = ParamFields {
              model  :: Maybe Flag
             --General
             ,nStart :: Maybe Int, nStop  :: Maybe Int, nStep  :: Maybe Int
             ,fStart :: Maybe Rational, fStop  :: Maybe Rational, fStep  :: Maybe Rational
             --TV-specific
             ,rStart :: Maybe Rational, rStop  :: Maybe Rational, rStep  :: Maybe Rational
             --Non-TV specific
             ,kStart :: Maybe Int, kStop  :: Maybe Int, kStep  :: Maybe Int
             --Small-world specific
             ,pStart :: Maybe Rational,pStop  :: Maybe Rational, pStep  :: Maybe Rational
             --Scale-free specific
             ,mStart :: Maybe Int, mStop  :: Maybe Int, mStep  :: Maybe Int
             --Kronecker/logit specific
             ,aStart :: Maybe Rational, aStop :: Maybe Rational, aStep :: Maybe Rational
             ,bStart :: Maybe Rational, bStop :: Maybe Rational, bStep :: Maybe Rational 
             --Socket specific
             ,sockets:: Maybe [Int]
             --accessible specific
             ,access :: Maybe AccessType
             --Kronecker specific
             ,initGraph :: Maybe InitGraph
             --Parameters with defaults for almost all models
             ,sigma  :: Int
             ,count  :: Int
             ,initT  :: Bool
             ,warn   :: String --maybe change to [String]
             } deriving (Show, Eq, Ord)



defaultParams = ParamFields Nothing
                         Nothing  Nothing  Nothing  --n
                         Nothing  Nothing  Nothing  --f
                         Nothing  Nothing  Nothing  --r
                         Nothing  Nothing  Nothing  --k
                         Nothing  Nothing  Nothing  --p
                         Nothing  Nothing  Nothing  --m
                         Nothing  Nothing  Nothing  --a
                         Nothing  Nothing  Nothing  --b
                         Nothing                    --sockets
                         Nothing                    --access
                         Nothing                    --initgraph
                         2 100 True ""


--These show the format of each type, and are used in comparisons by checkSanity to determine whether
--an input has a workable set of Justs and Nothings.
--Checksanity could also use something like this, with aprameter names
--tvPNames = ["n", "f", "r"]
--matchParams params pnames = 
--    ("n" in pnames) == (all.... 


exampleTV  = defaultParams {nStart = Just 0, nStop = Just 0, nStep = Just 0, 
                            fStart = Just 0, fStop = Just 0, fStep = Just 0, 
                            rStart = Just 0, rStop = Just 0, rStep = Just 0}

exampleSW  = ParamFields (Just SmallWorld)
                         (Just 0) (Just 0) (Just 0) --has the ns
                         (Just 0) (Just 0) (Just 0) --has the fs
                         Nothing  Nothing  Nothing  --no rs
                         (Just 0) (Just 0) (Just 0) --has the ks
                         (Just 0) (Just 0) (Just 0) --has the ps
                         Nothing  Nothing  Nothing  --no ms
                         Nothing  Nothing  Nothing  --no as
                         Nothing  Nothing  Nothing  --no bs
                         Nothing                    --no sockets
                         Nothing                    --no access
                         Nothing                    --no initgraph
                         0 0 True ""

exampleSF  = ParamFields (Just ScaleFree)
                         (Just 0) (Just 0) (Just 0) --has the ns
                         (Just 0) (Just 0) (Just 0) --has the fs
                         Nothing  Nothing  Nothing  --no rs
                         (Just 0) (Just 0) (Just 0) --has the ks
                         Nothing  Nothing  Nothing  --no ps
                         (Just 0) (Just 0) (Just 0) --has the ms
                         Nothing  Nothing  Nothing  --no as
                         Nothing  Nothing  Nothing  --no bs
                         Nothing                    --no sockets
                         Nothing                    --no access
                         Nothing                    --no initgraph
                         0 0 True ""

exampleNoKSF  = ParamFields (Just ScaleFree)
                         (Just 0) (Just 0) (Just 0) --has the ns
                         (Just 0) (Just 0) (Just 0) --has the fs
                         Nothing  Nothing  Nothing  --no rs
                         Nothing  Nothing  Nothing  --no ks
                         Nothing  Nothing  Nothing  --no ps
                         (Just 0) (Just 0) (Just 0) --has the ms
                         Nothing  Nothing  Nothing  --no as
                         Nothing  Nothing  Nothing  --no bs
                         Nothing                    --no sockets
                         Nothing                    --no access
                         Nothing                    --no initgraph
                         0 0 True ""

exampleSFB = ParamFields (Just ScaleFreeB)
                         (Just 0) (Just 0) (Just 0) --has the ns
                         (Just 0) (Just 0) (Just 0) --has the fs
                         Nothing  Nothing  Nothing  --no rs
                         (Just 0) (Just 0) (Just 0) --has the ms
                         Nothing  Nothing  Nothing  --no ps
                         Nothing  Nothing  Nothing  --no ms
                         Nothing  Nothing  Nothing  --no as
                         Nothing  Nothing  Nothing  --no bs
                         Nothing                    --no sockets
                         Nothing                    --no access
                         Nothing                    --no initgraph
                         0 0 True ""

exampleSoc = ParamFields (Just Socket)
                         Nothing  Nothing  Nothing  --no ns
                         (Just 0) (Just 0) (Just 0) --has the fs
                         Nothing  Nothing  Nothing  --no rs
                         Nothing  Nothing  Nothing  --no ks
                         Nothing  Nothing  Nothing  --no ps
                         Nothing  Nothing  Nothing  --no ms
                         Nothing  Nothing  Nothing  --no as
                         Nothing  Nothing  Nothing  --no bs
                         (Just [0])                 --has sockets
                         Nothing                    --no access
                         Nothing                    --no initgraph
                         0 0 True ""


exampleAccess  = ParamFields (Just Accessible) 
                         (Just 0) (Just 0) (Just 0) --has the ns
                         (Just 0) (Just 0) (Just 0) --has the fs
                         (Just 0) (Just 0) (Just 0) --has the rs
                         Nothing  Nothing  Nothing  --no ks
                         Nothing  Nothing  Nothing  --no ps
                         Nothing  Nothing  Nothing  --no ms
                         Nothing  Nothing  Nothing  --no as
                         Nothing  Nothing  Nothing  --no bs
                         Nothing                    --no sockets
                         (Just Access)              --no access
                         Nothing                    --no initgraph
                         0 0 True ""

exampleKronecker  = ParamFields (Just Kronecker) 
                         Nothing  Nothing  Nothing  --no ns
                         (Just 0) (Just 0) (Just 0) --has the fs
                         Nothing  Nothing  Nothing  --no rs
                         (Just 0) (Just 0) (Just 0) --has the ks
                         Nothing  Nothing  Nothing  --no ps
                         Nothing  Nothing  Nothing  --no ms
                         (Just 0) (Just 0) (Just 0) --has the as
                         (Just 0) (Just 0) (Just 0) --has the bs
                         Nothing                    --no sockets
                         Nothing                    --no access
                         (Just C2)                  --has initgraph
                         0 0 True ""

exampleVertexCopying  = ParamFields (Just VertexCopying) 
                         (Just 0) (Just 0) (Just 0) --has the ns
                         (Just 0) (Just 0) (Just 0) --has the fs
                         (Just 0) (Just 0) (Just 0) --has the rs
                         Nothing  Nothing  Nothing  --no ks
                         Nothing  Nothing  Nothing  --no ps
                         Nothing  Nothing  Nothing  --no ms
                         Nothing  Nothing  Nothing  --no as
                         (Just 0) (Just 0) (Just 0) --has the bs
                         Nothing                    --no sockets
                         Nothing                    --no access
                         Nothing                    --no initgraph
                         0 0 True ""

exampleLogit  = ParamFields (Just Logit) 
                         (Just 0) (Just 0) (Just 0) --has the ns
                         (Just 0) (Just 0) (Just 0) --has the fs
                         (Just 0) (Just 0) (Just 0) --has the rs
                         Nothing  Nothing  Nothing  --no ks
                         Nothing  Nothing  Nothing  --no ps
                         Nothing  Nothing  Nothing  --no ms
                         Nothing  Nothing  Nothing  --no as
                         Nothing  Nothing  Nothing  --no bs
                         Nothing                    --no sockets
                         Nothing                    --no access
                         Nothing                    --no initgraph
                         0 0 True ""

emptyParams = (ParamFields Nothing Nothing Nothing Nothing Nothing Nothing Nothing 
                           Nothing Nothing Nothing Nothing Nothing Nothing Nothing 
                           Nothing Nothing Nothing Nothing Nothing Nothing Nothing
                           Nothing Nothing Nothing Nothing Nothing Nothing Nothing 
                           2 1 True "the following lines are invalid: " )
                         
type TVParams  = (Bool,Int,Int,Rational,Rational)
type AccParams = (AccessType,Bool,Int,Int,Rational,Rational)
type SWParams  = (Int,Int,Int,Rational,Rational)
type SFParams  = (Bool,Int,Int,Int,Int,Rational)
type SFBParams = (Bool,Int,Int,Int,Rational)
type SocParams = (Bool,Int,[Int],Rational)
type KrParams  = (Bool,Int,InitGraph,Int,Rational,Rational,Rational)
type VCParams  = (Bool,Int,Int,Rational,Rational,Rational)
type LogParams = (Bool,Int,Int,Rational,Rational)

data ParamTuple = TVP TVParams | SWP SWParams | SFP SFParams | SFBP SFBParams | SocP SocParams | AccP AccParams | KrP KrParams | VCP VCParams | LogP LogParams

type Corpus = [(ParamTuple, [Automaton])]

data Flag = TV | SmallWorld | ScaleFree | ScaleFreeB | Socket | Accessible | Kronecker | VertexCopying | Logit deriving (Show,Eq,Ord)

--When calling from commandline, -o arg, --Option arg, and --Option=arg all work.
options :: [OptDescr String]
options = []
     {-[ Option ['b'] [] (ReqArg id "FILE") "gets file"
     , Option ['o'] [] (ReqArg id "FILE") "gets file"
     --maybe an option for a seed flag?
     ]-}

readRational :: String -> Rational
readRational str = fst.head $ Numeric.readFloat str

doubleOfRational :: Rational -> Double
doubleOfRational x = fromRational x

main = 
  do
    arguments  <- getArgs
    let filePaths = processArguments arguments
    (inh,tentODir) <- getHandles filePaths
    inpString <- hGetContents inh
    let inpLines = lines inpString 
    let params = insertDefaults $ getParams (map (splitOn "=" . stripAll) inpLines)
    generator <- getStdGen
    let (corpus,_) = if checkSanity params 
                     then MS.runState (corpusOfParams params) generator
                     else error $ "fails the sanity test"
                     --should be fail, but fail doesn't compile...?
    outDir <- if tentODir == "" 
              then directoryOfParams params 
              else return tentODir
    printCorpus outDir corpus
    hClose inh

directoryOfParams :: AutomataParams -> IO FilePath
--make sigmaStr and initStr show up when they're not default
directoryOfParams params = 
    let sm field  = fromMaybe "" $ (liftM show) (field params)
        smf field = fromMaybe "" $ (liftM sr) (field params)
        nStr = intercalate "-" ["n", sm nStart, "to", sm nStop, "by", sm nStep]
        fStr = intercalate "-" ["f", smf fStart, "to", smf fStop, "by", smf fStep]
        rStr = intercalate "-" ["r", smf rStart, "to", smf rStop, "by", smf rStep]
        kStr = intercalate "-" ["k", sm kStart, "to", sm kStop, "by", sm kStep]
        pStr = intercalate "-" ["p", smf pStart, "to", smf pStop, "by", smf pStep]
        mStr = intercalate "-" ["m", sm mStart, "to", sm mStop, "by", sm mStep]
        aStr = intercalate "-" ["a", smf aStart, "to", smf aStop, "by", smf aStep]
        bStr = intercalate "-" ["b", smf bStart, "to", smf bStop, "by", smf bStep]
        calcKr k = show $ krCalcN (fromMaybe C2 (initGraph params)) (fromMaybe 0 k)
        krNStr = intercalate "-" ["n", calcKr (kStart params), "to", calcKr (kStop params)]
        socStr = "soc-" ++ (tail $ init (replaceAll ',' '-' (sm sockets)))
        --initStr  = "initT-" ++ (show $ initTrans params) 
        --sigmaStr = "sigma-" ++ (show $ sigma params)
        countStr = "count-" ++ (show $ count params)
        accessStr = "access-" ++ (fromMaybe "" $ (liftM show) (access params))
        initStr = "G-" ++ (fromMaybe "" $ (liftM show) (initGraph params))
        
        paramStr = case model params of
                        Just TV         -> intercalate "_" ["TV", nStr, rStr, fStr, countStr]
                        Just SmallWorld -> intercalate "_" ["SW", nStr, kStr, pStr, fStr, countStr]
                        Just ScaleFree  -> intercalate "_" ["SF", nStr, mStr, kStr, fStr, countStr]
                        Just ScaleFreeB -> intercalate "_" ["SFB", nStr, kStr, fStr, countStr]
                        Just Socket     -> intercalate "_" ["Soc", socStr, countStr]
                        Just Accessible -> intercalate "_" ["Acc", accessStr, nStr, rStr, fStr, countStr]
                        Just Kronecker  -> intercalate "_" ["Kr", initStr, kStr, krNStr, aStr, bStr, fStr, countStr]
                        Just Logit      -> intercalate "_" ["Log", nStr, rStr, fStr, countStr]
                        Just VertexCopying -> intercalate "_" ["VC", nStr, bStr, rStr, fStr, countStr]
                        _ -> error (show params)
    in createNewDirectory paramStr

replaceAll key val (x:xs) = if x == key then val : replaceAll key val xs else key : replaceAll key val xs

printCorpus :: FilePath -> Corpus -> IO [()]
--Prints out the entire corpus to the out director.
printCorpus outDir corpus@((_,auts):_) = sequence $ map (printDataPoint outDir (length auts)) corpus
printCorpus outDir [] = fail "empty corpus"

printDataPoint :: FilePath -> Int -> (ParamTuple,[Automaton]) -> IO ()
--Prints out one of the data points, which is count automata tupled with the ParamTuple.
printDataPoint _ _ (_,[]) = return ()
printDataPoint outDir size (params,(aut:auts)) = 
	     let partialPath = outDir ++ "/" ++ (cleanParams params)
	     in do
		  outFile <- openFreshFile (partialPath ++ "--" ++ (show ((length auts) + 1)) 
					 ++ "-of-" ++ (show size))
		  hPutStr outFile (stringOfAutomata Raskin (Right aut))
		  hClose outFile
		  printDataPoint outDir size (params,auts)

sr a = (showFFloat Nothing (fromRational a)) ""

cleanParams :: ParamTuple -> String
--Turns a ParamTuple into a String suitable for filenames.
cleanParams (TVP (initTrans,sigma,n,r,f)) = 
             "TV-n-" ++ (show n) ++ "-r-" ++ (sr r) ++ "-f-" ++ (sr f)

cleanParams (SWP (sigma,n,k,p,f)) = 
             "SW-n-" ++ (show n) ++ "-k-" ++ (show k) ++ "-p-" ++ (sr p) ++ "-f-" ++ (sr f)

cleanParams (SFP (initTrans,sigma,n,m,k,f)) = 
             "SF-n-" ++ (show n) ++ "-m-" ++ (show m) ++ "-k-" ++ (show k) ++ "-f-" ++ (sr f)

cleanParams (SFBP (initTrans,sigma,n,k,f)) = 
             "SFB-n-" ++ (show n) ++ "-k-" ++ (show k) ++ "-f-" ++ (sr f)

cleanParams (SocP (initTrans,sigma,degrees,f)) = 
             "Soc-d-" ++ cleanD ++ "-f-" ++ (sr f)
             where
               cleanD = show (foldr (+) 0 degrees)

cleanParams (AccP (accessibility,initTrans,sigma,n,r,f)) =
             "Acc-n-" ++ (show n) ++ "-acc-" ++ (show accessibility) ++  "-r-" ++ (sr r) ++ "-f-" ++ (sr f)

cleanParams (KrP (initTrans,sigma,arr,k,a,b,f)) =
             "Kr-graph-" ++ (show arr) ++  "-k-" ++ (show k) ++ "-n-" ++ (show (krCalcN arr k)) ++ "-a-" ++ (sr a) ++ "-b-" ++ (sr b) ++ "-f-" ++ (sr f)

cleanParams (VCP (initTrans,sigma,n,b,r,f)) =
             "VC-n-" ++ (show n) ++ "-b-" ++ (sr b) ++ "-r-" ++ (sr r) ++ "-f-" ++ (sr f)

cleanParams (LogP (initTrans,sigma,n,r,f)) =
             "Log-m-" ++ (show n) ++ "-n-" ++ (show ((n*(n-1)) `div` 2))  ++ "-r-" ++ (sr r) ++ "-f-" ++ (sr f)

openFreshFile :: FilePath -> IO Handle
--HAS RACE CONDITIONS.
openFreshFile path  = 
    let aux path count =
             do
               let path' = pathWithNumber path count ".aut"
               exists <- doesFileExist path'
               if exists
               then aux path (count+1)
               else openFile path' WriteMode
    in aux path 0

createNewDirectory path = 
    let aux path count =
             do
               let path' = pathWithNumber path count ""
               exists <- doesFileExist path'
               if exists
               then aux path (count+1)
               else createDirectory path' >> return path'
    in aux path 0

pathWithNumber path count ext = 
        if count == 0 
        then path ++ ext
        else path ++ "-" ++ (show count) ++ ext

corpusOfParams :: AutomataParams -> RandomState Corpus
--Generates a corpus of the right type corresponding to the AutomataParams object passed in.
corpusOfParams params = 
             case model params of
               Just TV         -> do
                                    let paramTuple = tvParams params
                                    corpus <- makeCorpus (count params) (uncurryN tabakovVardiAutomaton) paramTuple
                                    return $  map (\(x,y) -> (TVP x,y)) corpus
               Just SmallWorld -> do
                                    let paramTuple = swParams params
                                    corpus <- makeCorpus (count params) (uncurryN smallWorldAutomaton) paramTuple
                                    return $ map (\(x,y) -> (SWP x,y)) corpus
               Just ScaleFree  -> do
                                    let paramTuple = sfParams params
                                    corpus <- makeCorpus (count params) (uncurryN scaleFreeAutomaton) paramTuple
                                    return $ map (\(x,y) -> (SFP x,y)) corpus
               Just ScaleFreeB -> do
                                    let paramTuple = sfbParams params
                                    corpus <- makeCorpus (count params) (uncurryN scaleFreeBAutomaton) paramTuple 
                                    return $ map (\(x,y) -> (SFBP x,y)) corpus
               Just Socket     -> do
                                    let paramTuple = socParams params
                                    corpus <- makeCorpus (count params) (uncurryN socketAutomaton) paramTuple
                                    return $ map (\(x,y) -> (SocP x,y)) corpus
               Just Accessible -> do
                                    let paramTuple = accParams params
                                    corpus <- makeCorpus (count params) (uncurryN biAccessibleAutomaton) paramTuple
                                    return $ map (\(x,y) -> (AccP x,y)) corpus
               Just Kronecker  -> do
                                    let paramTuple = krParams params
                                    corpus <- makeCorpus (count params) (uncurryN kroneckerByInit) paramTuple
                                    return $ map (\(x,y) -> (KrP x,y)) corpus
               Just Logit      -> do
                                    let paramTuple = logitParams params
                                    corpus <- makeCorpus (count params) (uncurryN cleanedLogitAutomaton) paramTuple
                                    return $ map (\(x,y) -> (LogP x,y)) corpus
               Just VertexCopying -> do
                                       let paramTuple = vcParams params
                                       corpus <- makeCorpus (count params) (uncurryN vertexCopyingAutomaton) paramTuple
                                       return $ map (\(x,y) -> (VCP x,y)) corpus
                                       
tripleToList :: (Enum a, Num a, Ord a) => Maybe a -> Maybe a -> Maybe a -> Maybe [a]
--Using a start, stop, and step, generates the corresponding list using ranges. Overrides Haskell's normal behavior for
--doubles in ranges; it's now less forgiving for imprecision that pushes a number past the stop value when it shouldn't.
tripleToList start stop step = 
             do
               s1 <- start
               s2 <- stop
               s3 <- step
               return [x | x <- [s1,s1+s3..s2], x <= s2]

--This set of functions converts an AutomataParams to the list of tuples that makeCorpus takes for the appropriate automaton.
--If one of the required parameters is a Nothing, then it will return an empty list, and makeCorpus will generate an empty
--list as well.
--Needs to have some edits to deal with the current method of using ifs in list comprehensions to give warnings.
tvParams :: AutomataParams -> [TVParams]
tvParams params = fromMaybe [] 
                      (do
                        ns <- tripleToList (nStart params) (nStop params) (nStep params)
                        rs <- tripleToList (rStart params) (rStop params) (rStep params)
                        fs <- tripleToList (fStart params) (fStop params) (fStep params)
                        return [(initT params,sigma params,n,r,f) | n <- ns, r <- rs, f <- fs, (floor ((fromIntegral n) * r)) < n*n])

swParams :: AutomataParams -> [SWParams]
swParams params = fromMaybe []
                      (do
                        ns <- tripleToList (nStart params) (nStop params) (nStep params)
                        ks <- tripleToList (kStart params) (kStop params) (kStep params)
                        ps <- tripleToList (pStart params) (pStop params) (pStep params)
                        fs <- tripleToList (fStart params) (fStop params) (fStep params)
                        return [(sigma params,n,k,p,f) | n <- ns, k <- ks, p <- ps, f <- fs, k < (n `div` 2)])

sfParams :: AutomataParams -> [SFParams]
--Has a kludge to get K to default to M.
sfParams params = fromMaybe []
                      (do
                        ns <- tripleToList (nStart params) (nStop params) (nStep params)
                        ms <- tripleToList (mStart params) (mStop params) (mStep params)
                        fs <- tripleToList (fStart params) (fStop params) (fStep params)
                        if (kStart params) == Nothing 
                        then return [(initT params,sigma params,n,m,m,f) | n <- ns, m <- ms, f <- fs]
                        else do
                               ks <- tripleToList (kStart params) (kStop params) (kStep params)
                               return [(initT params,sigma params,n,m,k,f) | n <- ns, m <- ms, k <- ks, f <- fs, k <= 2*m]) 
                  

sfbParams :: AutomataParams -> [SFBParams]
sfbParams params = fromMaybe []
                       (do
                         ns <- tripleToList (nStart params) (nStop params) (nStep params)
                         ks <- tripleToList (kStart params) (kStop params) (kStep params)
                         fs <- tripleToList (fStart params) (fStop params) (fStep params)
                         return [(initT params,sigma params,n,k,f) | n <- ns, k <- ks, f <- fs, k <= n])

socParams :: AutomataParams -> [SocParams]
socParams params = fromMaybe [] 
                       (do
                         fs <- tripleToList (fStart params) (fStop params) (fStep params)
                         socs <- sockets params
                         return [(initT params,sigma params,socs,f) | f <- fs])


accParams :: AutomataParams -> [AccParams]
accParams params = fromMaybe [] 
                      (do
                        ns <- tripleToList (nStart params) (nStop params) (nStep params)
                        rs <- tripleToList (rStart params) (rStop params) (rStep params)
                        fs <- tripleToList (fStart params) (fStop params) (fStep params)
                        p  <- access params
                        return [(p,initT params,sigma params,n,r,f) | n <- ns, r <- rs, f <- fs, (floor ((fromIntegral n) * r)) < n*n])

krParams :: AutomataParams -> [KrParams]
krParams params = fromMaybe []
                      (do
                        ks <- tripleToList (kStart params) (kStop params) (kStep params)
                        as <- tripleToList (aStart params) (aStop params) (aStep params)
                        bs <- tripleToList (bStart params) (bStop params) (bStep params)
                        fs <- tripleToList (fStart params) (fStop params) (fStep params)
                        i  <- initGraph params
                        return [(initT params,sigma params,i,k,a,b,f) | k <- ks, a <- as, b <- bs, f <- fs])

vcParams :: AutomataParams -> [VCParams]
vcParams params = fromMaybe []
                      (do
                        ns <- tripleToList (nStart params) (nStop params) (nStep params)
                        bs <- tripleToList (bStart params) (bStop params) (bStep params)
                        rs <- tripleToList (rStart params) (rStop params) (rStep params)
                        fs <- tripleToList (fStart params) (fStop params) (fStep params)
                        return [(initT params,sigma params,n,b,r,f) | n <- ns, b <- bs, r <- rs, f <- fs])
                        
logitParams :: AutomataParams -> [LogParams]
logitParams params = fromMaybe []
                      (do
                        ns <- tripleToList (nStart params) (nStop params) (nStep params)
                        rs <- tripleToList (rStart params) (rStop params) (rStep params)
                        fs <- tripleToList (fStart params) (fStop params) (fStep params)
                        return [(initT params,sigma params,n,r,f) | n <- ns, r <- rs, f <- fs, (floor ((fromIntegral n) * r)) < n*n])

processArguments :: [String] -> [String]
--turns the arguments into something useful
processArguments args =
    let (o,f,errs) = getOpt Permute options args
    in f

getHandles :: [FilePath] -> IO (Handle,FilePath)
--pulls out a handle from the file paths
getHandles [] = fail "corpus creation requires an input file"
getHandles (ipath:[]) = 
             do
               inh <- openFile ipath ReadMode
               return (inh,"")
getHandles (ipath:odir:_) =
             do
               inh <- openFile ipath ReadMode
               return (inh,odir)

stripAll :: String -> String
--removes all whitespace from a string (including middle! Be careful!)
stripAll str = filter (not . isSpace) str

getParams :: [[String]] -> AutomataParams
--inserts the cfg file data into the AutomataParams
getParams [] = emptyParams
getParams ((id:dat:_):xs) =
             case id of
               "nStart"     -> (getParams xs) { nStart = Just (read dat)         }
               "nStop"      -> (getParams xs) { nStop  = Just (read dat)         }
               "nStep"      -> (getParams xs) { nStep  = Just (read dat)         }
               "fStart"     -> (getParams xs) { fStart = Just (readRational dat) }
               "fStop"      -> (getParams xs) { fStop  = Just (readRational dat) }
               "fStep"      -> (getParams xs) { fStep  = Just (readRational dat) }
               "rStart"     -> (getParams xs) { rStart = Just (readRational dat) }
               "rStop"      -> (getParams xs) { rStop  = Just (readRational dat) }
               "rStep"      -> (getParams xs) { rStep  = Just (readRational dat) }
               "mStart"     -> (getParams xs) { mStart = Just (read dat)         }
               "mStop"      -> (getParams xs) { mStop  = Just (read dat)         }
               "mStep"      -> (getParams xs) { mStep  = Just (read dat)         }
               "kStart"     -> (getParams xs) { kStart = Just (read dat)         }
               "kStop"      -> (getParams xs) { kStop  = Just (read dat)         }
               "kStep"      -> (getParams xs) { kStep  = Just (read dat)         }
               "pStart"     -> (getParams xs) { pStart = Just (readRational dat) }
               "pStop"      -> (getParams xs) { pStop  = Just (readRational dat) }
               "pStep"      -> (getParams xs) { pStep  = Just (readRational dat) }
               "aStart"     -> (getParams xs) { aStart = Just (readRational dat) }
               "aStop"      -> (getParams xs) { aStop  = Just (readRational dat) }
               "aStep"      -> (getParams xs) { aStep  = Just (readRational dat) }
               "bStart"     -> (getParams xs) { bStart = Just (readRational dat) }
               "bStop"      -> (getParams xs) { bStop  = Just (readRational dat) }
               "bStep"      -> (getParams xs) { bStep  = Just (readRational dat) }
               "sockets"    -> (getParams xs) { sockets= Just (read dat)         }
               "access"     -> (getParams xs) { access = Just (read dat)         }
               "initGraph"  -> (getParams xs) { initGraph = Just (read dat)      }
               "sigma"      -> (getParams xs) { sigma  = read dat                }
               "count"      -> (getParams xs) { count  = read dat                }
               "initT"      -> (getParams xs) { initT  = read dat                }
               _            -> (getParams xs) { warn   = dat                     } --only gives one warning
getParams ((flag:[]):xs) =
             case flag of
               "TV"         -> (getParams xs) { model = Just TV         }
               "SmallWorld" -> (getParams xs) { model = Just SmallWorld }
               "ScaleFree"  -> (getParams xs) { model = Just ScaleFree  }
               "ScaleFreeB" -> (getParams xs) { model = Just ScaleFreeB }
               "Socket"     -> (getParams xs) { model = Just Socket     }
               "Accessible" -> (getParams xs) { model = Just Accessible }
               "Kronecker"  -> (getParams xs) { model = Just Kronecker  }
               "Logit"      -> (getParams xs) { model = Just Logit      }
               "VertexCopying" -> (getParams xs) { model = Just VertexCopying }
               _            -> (getParams xs) { warn  = flag            }
getParams (_:xs) = getParams xs

checkSanity :: AutomataParams -> Bool
--makes sure that the parameters inputted match an acceptable set
checkSanity params = any (approxMatch params) [exampleTV, exampleSW, exampleSF, exampleNoKSF, exampleSFB, exampleSoc, exampleAccess,exampleKronecker,exampleVertexCopying,exampleLogit]

approxMatch params1 params2 = 
--Ensures that a parameter set has the appropriate Justs and Nothings.
    let intFields = [nStart, nStop, nStep, mStart, mStop, mStep, kStart, kStop, kStep]
        rationalFields = [fStart, fStop, fStep, rStart, rStop, rStep, pStart, pStop, pStep,aStart,aStop,aStep,bStart,bStop,bStep]
        checkField field = case (field params1, field params2) of
                            (Just _, Just _) -> True
                            (Nothing, Nothing) -> True
                            _ -> False
    in checkField model && (all checkField intFields) && (all checkField rationalFields) && 
       checkField sockets && checkField access && checkField initGraph

insertDefaults :: AutomataParams -> AutomataParams
--fiddle with this later
insertDefaults params = params
