module RandomState where

import Data.List
import Data.Ratio
import Control.Monad.State as MS
import System.Random
import Automata

type RandomState a = MS.State StdGen a
--Conceptually, a RandomState is a value that depends on a randomly generated number. Any RandomState a will, upon being run with runState and a StdGen,
--generate a tuple (a,s) where s is the new StdGen and a is a value. The value can be used in computations, the StdGen keeps track of your generator 
--in a sequence of random computations instead of having to do it manually.

makeCorpus :: Int -> (a -> RandomState b) -> [a] -> RandomState [(a,[b])]
--Automata take tupled parameters. The generator produces Int random-state automata of a type determined by the function for each of the parameter-tuples 
--in the list, and returns them as one large list of tuples of the parameters used to generate a particular automaton, and the automaton.
makeCorpus _ _ [] = return []
makeCorpus n func ps = 
        let makePoint p = 
               do
                 autList <- sequence $ replicate n (func p) 
                 return (p,autList)
        in sequence (map makePoint ps)

getLabeledSubset :: (Eq a, Eq b) => [a] -> [b] -> Int -> RandomState [(a,b)]
--Randomly matches count vals with labels.
getLabeledSubset labels vals count =
         let aux _ _ curr 0 = return curr
             aux l v curr i =
               do
                 lIndex <- getRandomR (0,(length l)-1)
                 vIndex <- getRandomR (0,(length v)-1)
                 let x = (l !! lIndex,v !! vIndex)
                 if x `elem` curr
                 then aux l v curr i
                 else aux l v (x:curr) (i-1)
         in aux labels vals [] count

--same as GetRandomMultisubset, but nubs the list so as to return a proper set
getRandomSubset set n = getRandomMultisubset (nub set) n

getRandomMultisubset :: (Eq a) => [a] -> Int -> RandomState [a]
--getRandomMultisubset takes a list and an integer n, and generates a random multisubset of the list of size n.
getRandomMultisubset _ 0 = return []
getRandomMultisubset [] _ = fail "Not enough elements for getRandomSubset"
getRandomMultisubset set n = 
          do
            index <- getRandomR (0,(length set) - 1)
            let res = set !! index 
            restElems <- getRandomMultisubset (delete res set) (n - 1)
            return $ res : restElems

getRandomElem :: (Eq a) => [a] -> RandomState a
--Unsafe against empty lists. Chooses a random element from a list.
getRandomElem lst =
          do
            index <- getRandomR (0,(length lst)-1)
            return $ lst !! index


getRandomR :: (Random a) => (a,a) -> RandomState a
--getRandomR takes a range, and returns a monadic random number inside the range.
getRandomR range = 
          do
            generator <- MS.get
            let (res,generator') = randomR range generator
            MS.put generator'
            return res
