{-# LANGUAGE ScopedTypeVariables #-}

module ScaleFreeAutomata where

import Automata
import RandomState
import Data.List
import Data.Ratio
import qualified Control.Monad.State as MS

default (Int)

socketAutomaton :: Bool -> Int -> [Int] -> Rational -> RandomState Automaton
--initTrans forces an outbound transition on each character from the start state.
--Degree distribution is /not/ guaranteed to be realized because there may be duplicate edges.
--Takes sigma = number of characters, degrees = a list of ints where index 0 holds the number of nodes with degree 0, index
--1 holds the number of nodes with degree 1, etc, f = final state density to generate an automaton of a given distribution.
socketAutomaton initTrans sigma degrees f =
         let size = sum degrees
         in do
              graphs <- sequence $ replicate sigma (socketGraph initTrans degrees size)
              finStates <- getRandomSubset [0..size-1] (floor (f * (fromIntegral size)))
              return $ automataCompose [0] graphs finStates

socketGraph :: Bool -> [Int] -> Int -> RandomState Graph
--Takes a list of degrees and the number of nodes and generates a random graph with that degree distribution.
socketGraph initTrans degrees n =
         do
           let socketList = degreesToSockets degrees 0 0
           edges <- if initTrans
                    then initEdge socketList
                    else genEdges socketList
           return (n,edges)
         where
           genEdges :: [Int] -> RandomState [(Int,Int)]
           genEdges (x:[]) = fail "impossible distribution, must be an even number of sockets"
           genEdges [] = return []
           genEdges socketList =
                     do
                       (u:v:[]) <- getRandomSubset socketList 2
                       let edge = (u,v)
                       restEdges <- genEdges (delete u $ delete v socketList)
                       return $ edge : restEdges
           degreesToSockets :: [Int] -> Int -> Int -> [Int]
           degreesToSockets [] _ _ = []
           degreesToSockets (x:xs) currDegree currState = 
                    let thisDegree = concat $ replicate currDegree [currState..currState+x-1]
                    in thisDegree ++ (degreesToSockets xs (currDegree + 1) (currState + x))
           initEdge :: [Int] -> RandomState [(Int,Int)]
           initEdge socketList =
                     do
                       v <- getRandomElem socketList
                       let edge = (0,v)
                       restEdges <- genEdges (delete 0 $ delete v socketList)
                       return $ edge : restEdges

--make this not allow duplicate edges
scaleFreeAutomaton :: Bool -> Int -> Int -> Int -> Int -> Rational -> RandomState Automaton
--initTrans forces an outbound transition on each character from the start state.
--Takes sigma = number of characters, n = number of timesteps, m = starting size, r = edges per timestep, f = final state 
--density to generate a scale-free automaton.
scaleFreeAutomaton initTrans sigma n m k f =
         do
           graphs <- sequence $ replicate sigma (scaleFreeGraph initTrans n m k)
           finStates <- getRandomSubset [0..n-1] (floor (f * (fromIntegral n)))
           return $ automataCompose [0] graphs finStates



scaleFreeGraph :: Bool -> Int -> Int -> Int -> RandomState Graph
--Takes n = final size, m = starting size, k = edges per timestep, and generates a scale-free graph.
scaleFreeGraph initTrans n m k =
         do
           edges <- if not initTrans
                    then genEdges (n-m) m [] [0..m-1]
                    else ensureInitEdge (genEdges (n-m) m [] [0..m-1])
           return (n,edges)
         where
           genEdges :: Int -> Int -> [(Int,Int)] -> [Int] -> RandomState [(Int,Int)]
           genEdges 0 _ currEdges _ = return currEdges
           genEdges count currState currEdges probList =
                     do
                       labeledVs <- getLabeledSubset [True,False] probList k
                       let vs = map (\(_,x) -> x) labeledVs
                       let newEdges = map (genEdge currState) labeledVs
                       let probList' = (replicate (k+1) currState) ++ vs ++ probList 
                       genEdges (count-1) (currState+1) (newEdges ++ currEdges) probList'
           genEdge currState (l,v) =
                     if l
                     then (currState,v)
                     else (v,currState)
           ensureInitEdge :: RandomState [(Int, Int)] -> RandomState [(Int, Int)]
           ensureInitEdge edges =
                     do
                       es <- edges
                       v <- getRandomR (0,n-1)
                       if not (any (\x -> (fst x) == 0) es)
                       then return ((0,v):es)
                       else return es

scaleFreeBAutomaton :: Bool -> Int -> Int -> Int -> Rational -> RandomState Automaton
--initTrans forces an outbound transition on each character from the start state.
--Takes sigma = number of characters, n = number of states, k = average degree of nodes, f = final state density to generate a scale-
--free automaton of constant size, rather than one that grows.
scaleFreeBAutomaton initTrans sigma n k f =
         do
           graphs <- sequence $ replicate sigma (scaleFreeBGraph initTrans n k)
           finStates <- getRandomSubset [0..n-1] (floor (f * (fromIntegral n)))
           return $ automataCompose [0] graphs finStates

scaleFreeBGraph :: Bool -> Int -> Int -> RandomState Graph
--Takes n = number of nodes, k = average degree of nodes to generate a scale-free graph.
scaleFreeBGraph initTrans n k =
         do
           if k > n then fail "n*k must be less than n^2, otherwise too many edges" else return ()
           edges <- genEdges initTrans (n*k) [] [0..n-1]
           return (n,edges)
         where
           genEdges :: Bool -> Int -> [(Int,Int)] -> [Int] -> RandomState [(Int,Int)]
           genEdges _ 0 currEdges _ = return currEdges
           genEdges True 1 currEdges probList = 
                     if any (\x -> (fst x) == 0) currEdges
                     then genEdges False 1 currEdges probList
                     else do
                            v <- getRandomElem probList
                            return $ (0,v) : currEdges
           genEdges init count currEdges probList = 
                     do
                       u <- getRandomR (0,n-1)
                       p <- getRandomR (0,(length probList) -1)
                       v <- getRandomElem probList
                       dir::Int <- getRandomR (1,2)
                       let newEdge = if dir == 1 
                                     then (u,v) 
                                     else (v,u) 
                       if newEdge `elem` currEdges 
                       then genEdges init count currEdges probList 
                       else genEdges init (count-1) (newEdge : currEdges) (u : v : probList)
