module SmallWorldAutomata where

import Automata
import RandomState
import Data.List
import Data.Ratio
import qualified Control.Monad.State as MS

--This set of functions generates a small world automaton by the method of rewiring existing edges instead of addingedges.

smallWorldAutomaton :: Int -> Int -> Int -> Rational -> Rational -> RandomState Automaton
--Taking sigma = number of characters, n = size, k = degree (nodes have an out-degree of 2k), p = rewire density, and f = accepting density
--generates a small-world automaton.
smallWorldAutomaton sigma n k p f =
        do
          trans <- sequence (replicate sigma (smallWorldGraph n k p))
          finStates <- getRandomSubset [0..n-1] (floor (f * (fromIntegral n)))
          return (automataCompose [0] trans finStates)


dirRingLattice :: Int -> Int -> Graph
--dirRingLattice takes a number of nodes n and a degree k; each node has 2k outgoing edges
--Disallows duplicates; if k is high enough that an edge can appear twice, it only appears once.
dirRingLattice n k = (n,nub [(u,v) | u <- nodes, v <- (delete u (destList [u-k..u+k]))])
                  where
                      --n is the number of nodes, but node numbering starts at 0
                      nodes = [0..n-1]
                      --destList wraps edges around so that instead of going to a nonexistent node, it acts as if we'd done mathematical mod to it
                      destList (x:xs) = if x < 0
                                        then destList ((n+x):xs)
                                        else if x >= n
                                             then destList ((x - n):xs)
                                             else x : destList xs
                      destList [] = [] 

smallWorldRewire :: Graph -> Rational -> RandomState Graph
--smallWorldRewire, given a graph (which should be a directed ring lattice) and a Rational between 0 and 1 inclusive p, takes n*2k*p edges and rewires
--them to create shortcuts. It does not allow duplicate edges, generating a new edge instead if one comes up. All chosen edges to be rewired are
--removed from the edge set before any rewired edges are added back in.
smallWorldRewire (n,edges) p =
                  do
                    let numRewire = floor $ (fromIntegral $ length edges) * p
                    subset <- getRandomSubset edges numRewire
                    newEdges <- rewireSet subset (edges \\ subset)
                    return (n,newEdges)
                  where
                    rewire :: (Int,Int) -> [(Int,Int)] -> RandomState [(Int,Int)]
		    --rewire takes the edge to be rewired, and a list of all current edges. It rewires it so that a (u,v) instead becomes (u,v'), with a v' 
            --chosen uniformly at random, without allowing duplicates. It returns a list for the sake of simplifying rewireSet.
                    rewire edge@(u,_) currEdges = 
                               do
                                 v <- getRandomR (0,n-1)
                                 let newEdge = (u,v)
                                 if newEdge `elem` currEdges
                                 then rewire edge currEdges
                                 else return $ newEdge : currEdges
                    rewireSet :: [(Int,Int)] -> [(Int,Int)] -> RandomState [(Int,Int)]
		    --rewireSet takes a list of edges to be rewired, and a list of all other edges to ensure no duplicates, and returns a [Edges] of the 
            --combined rewired edges and other edges.
                    rewireSet [] currEdges = return currEdges
                    rewireSet (e:es) currEdges = 
                               do 
                                 newEdges <- rewire e currEdges
                                 rewireSet es newEdges

smallWorldGraph :: Int -> Int -> Rational -> RandomState Graph
--smallWorldGraph n k p creates a small-world graph with n nodes, degree k, and chance p of rewiring.
smallWorldGraph n k p = let lattice = dirRingLattice n k
                        in smallWorldRewire lattice p
