import System.Environment
import System.IO
import System.Random
import Data.List
import Data.List.Split
import Data.Maybe
import Data.Char
import Debug.Trace
import Data.Tuple.Curry
import Control.Monad.State as MS
import SmallWorldAutomata
import TabakovVardi
import Automata
import RandomState
import ScaleFreeAutomata
import Kronecker
import VertexCopying
import Logit
import System.Console.GetOpt

validateEdges :: Int -> [(Int,Int)] -> Bool
--Checks that every u and every v in the edges [(u,v)] are below the input value
validateEdges _ [] = True
validateEdges n ((u,v):es)
                   | u >= n = False
                   | v >= n = False
                   | u < 0  = False
                   | v < 0  = False
                   | otherwise = validateEdges n es

checkDuplicates :: (Eq a) => [a] -> Bool
--Returns false if the argument contains a duplicate
checkDuplicates lst = checkDuplicates' lst []
                   where
                     checkDuplicates' :: (Eq a) => [a] -> [a] -> Bool
                     checkDuplicates' [] _ = True
                     checkDuplicates' (x:xs) curr =
                             if x `elem` curr
                             then False
                             else checkDuplicates' xs (x : curr)

testTabakovVardiGraph :: Int -> Double -> Graph -> Bool
--takes a TV graph and makes sure it has the values that should be produced by its n and r
testTabakovVardiGraph n r (size,edges) = size == n &&
                                         length edges == floor ((fromIntegral n) * r) &&
                                         validateEdges n edges &&
                                         checkDuplicates edges

testSmallWorldGraphA :: Int -> Int -> Graph -> Bool
--takes either a small world graph or a ring lattice and makes sure that it has the values that should be produced by its r
--n and k.
testSmallWorldGraphA n k (size,edges) = size == n && 
                                        length edges == (n * 2 * k) &&
                                        validateEdges n edges &&
                                        checkDuplicates edges --also could change to edges == nub edges

testSmallWorldGraphB :: Graph -> Graph -> Bool
--tests that the small world graph has size properties it should keep from the lattice it's created from
testSmallWorldGraphB (smSize,smEdges) (latSize,latEdges) = length smEdges == length latEdges && 
                                                           latSize == smSize

testScaleFreeBGraph :: Int -> Int -> Graph -> Bool
--takes a scale-freeB graph and makes sure it has values that should be produced by its n and m.
testScaleFreeBGraph n m (size,edges) = size == n &&
                                       length edges == (n * m) &&
                                       validateEdges n edges &&
                                       checkDuplicates edges

testScaleFreeGraph :: Int -> Int -> Int -> Graph -> Bool
--takes a normal scale-free graph and makes sure it has values that should be produced by its n, m, and r.
testScaleFreeGraph n m r (size, edges) = size == n &&
                                         length edges == ((n-m) * r) || length edges == ((n-m) * r + 1) &&
                                         validateEdges n edges &&
                                         checkDuplicates edges

testSocketGraph :: [Int] -> Int -> Graph -> Bool
--takes a socket graph and makes sure it has values that should be produced by its n, m, and r.
testSocketGraph degrees n (size, edges) = size == n &&
                                          length edges == ((length degrees) `div` 2) &&
                                          validateEdges n edges

testVertexCopyingGraph :: Int -> Double -> Graph -> Bool
--takes a VC graph and makes sure it has the values that should be produced by its n and r
testVertexCopyingGraph n r (size,edges) = size == n &&
                                         length edges == floor ((fromIntegral n) * r) &&
                                         validateEdges n edges &&
                                         checkDuplicates edges

testLogitGraph :: Int -> Double -> (PGraph LogitState) -> Bool
--takes a logit graph, makes sure it produces values that should be produced by its n and r
testLogitGraph n r (size,edges) = let states = [(x,y) | x <- [0..n-1], y <- [x..n-1]]
                                  in size == states &&
                                     length edges == floor ((fromIntegral (length states)) * r) &&
                                     validateLogitEdges states edges &&
                                     checkDuplicates edges

validateLogitEdges :: [(Int,Int)] -> [((Int,Int),(Int,Int))] -> Bool
--Checks that all the edges contain valid states, and are between valid states
validateLogitEdges _ [] = True
validateLogitEdges states ((u@(u1,u2),v@((v1,v2))):restOfEdges) = 
                                  u `elem` states &&
                                  v `elem` states &&
                                  (u1 == v1 || u1 == v2 || u2 == v1 || u2 == v2) &&
                                  validateLogitEdges states restOfEdges
                                  
--tests a composed automaton against one of the graphs making it up, to ensure that it has the right amount of transitions,
--size, etc
testComposedAutomata (Automaton sigma states startStates rho finStates) (size,edges) = 
                                         (length rho) `div` sigma == length edges &&
                                         length states == size &&
                                         length startStates <= size &&
                                         length finStates <= size
